'''
Fortress is a Typed ODM for Deta's Base NoSQL Database.

The wrapper exists to aid in testing and code reliability.
It uses dataclass objects to map to Bases. This allows
one to query the bases with these models. It also includes
a verbose testing suite to aid in development.
'''

# Exports
from .elysium import Elysium
